/* Drawing functions for memory frame buffers. A frame buffer is an instance
of a CMem class and can be shown on screen by a CScreen instance.
Use CDraw to draw in frame buffer: polygons, pixels, ... or clear it.

FEATURES:
- draw polygons with one color (flat polygons) or with texture. The polygons
  can have any number of vertexes.
- is possible draw some types of polygons making clipping through a edge
  buffer. This edge buffer is used to define the left and right limits of X
  that can be drawn per scanline, providing the clipping. The buffer can
  be defined as the edges of a clipping polygon using the metod
  CDraw::mSetClipEdgeBuffer and can be reseted by CDraw::mResetClipEdgeBuffer
  The clip edge buffer must be an array of type tagInterval with Y resolution
  size.

TIPS:
- you can do "enviroment mapping" using a texture as the frame buffer.
- you can save an image in RAW file format saving the frame buffer to a file.
- texture width and height must not be equal.
- you can use textures of any size to draw texture mapped polygons but:
  64x64 and 256x256 textures are draw very faster than textures with
  other sizes. Because acording the texture size the line draw loop
  must change.
- the reference system is the same as the video:    (0,0)
                                                      *----->Y
                                                      |
                                                      |
                                                      |
                                                      VX
WARNING:
- polygons to be drawn must be convex
*/



#include <string.h>
#include <math.h>

#include "defs.h"

#include "mem.h"

#include "string.h"

#include "texture.h"

#include "lighttab.h"

#include "draw.h"



// compute increase rate for x
#define INCREASEX( XI, YI, XF, YF ) ( (float) ( ( XF ) - ( XI ) ) / ( ( YF ) - ( YI ) ) )



/* constructor - init empty object
**********************************************************/
CDraw::CDraw ()
{
   m_nMidX = m_nMidY = m_nMaxX = m_nMaxY = m_nResX = m_nResY = m_nSize = 0;
   m_poFrBuff = NULL;
   m_panBackwardX = m_panForwardX = NULL;
   m_pastBackwardTex = m_pastForwardTex = NULL;
   m_nTexWidth = m_nTexHeight = 0;
   m_poLightTable = NULL;
}



/* constructor - init with an specific resolution.
**********************************************************/
CDraw::CDraw ( CLightTable& oLgT, INT nResX, INT nResY, CMem& oFrBuff )
{
   m_nMidX = m_nMidY = m_nMaxX = m_nMaxY = m_nResX = m_nResY = m_nSize = 0;
   m_poFrBuff = NULL;
   m_panBackwardX = m_panForwardX = NULL;
   m_pastBackwardTex = m_pastForwardTex = NULL;
   m_nTexWidth = m_nTexHeight = 0;
   m_poLightTable = NULL;

   mCreate ( oLgT, nResX, nResY, oFrBuff );
}



/* destructor - empty object
**********************************************************/
CDraw::~CDraw ()
{
   mDestroy ();
}



/* Set object member vriables. If frame buffer is
not bigger enought to this resolution realocate it.
return TRUE - ok \ FALSE - error
**********************************************************/
BOOL CDraw::mCreate ( CLightTable& oLgT, INT nResX, INT nResY, CMem& oFrBuff )
{
   mDestroy ();

   m_nResX = nResX;
   m_nResY = nResY;
   m_poFrBuff = &oFrBuff;
   m_nSize = m_nResX * m_nResY;
   m_nMaxX = m_nResX - 1;
   m_nMaxY = m_nResY - 1;
   m_nMidX = m_nResX >> 1;  //    >> 1 is equal / 2
   m_nMidY = m_nResY >> 1;
   m_nTexWidth = m_nTexHeight = 0;
   m_poLightTable = &oLgT;

   // try alocate memory for edge buffer for forward scan
   if ( ( m_panForwardX = new INT [ m_nResY ] ) == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // try alocate memory for edge buffer for backward scan
   if ( ( m_panBackwardX = new INT [ m_nResY ] ) == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // try alocate memory for edge buffer for forward scan in texture
   if ( ( m_pastForwardTex = new tagXY [ m_nResY ] ) == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // try alocate memory for edge buffer for backward scan
   if ( ( m_pastBackwardTex = new tagXY [ m_nResY ] ) == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // verify if memory object is not bigger enought to this resolution
   if ( (INT) m_poFrBuff->m_dwSize < m_nResX * m_nResY )
      // try alocate more memory
      if ( ! m_poFrBuff->mRealocate ( m_nResX * m_nResY ) )
         return FALSE;

   return TRUE;
}



/* Empty member variables (memory desalocation)
**********************************************************/
void CDraw::mDestroy ()
{
   // verify if object have some value in variables
   if ( m_poFrBuff != NULL ) {
      m_nMidX = m_nMidY = m_nResX = m_nResY = m_nSize = 0;
      m_nTexWidth = m_nTexHeight = 0;
      m_poFrBuff = NULL;
      m_poLightTable = NULL;
   }

   // verify if object have some value in variable
   if ( m_poLightTable != NULL )
      m_poLightTable = NULL;

   // verify if edge buffer for forward scan is alocated, if yes desalocate
   if ( m_panForwardX != NULL ) {
      delete m_panForwardX;
      m_panForwardX = NULL;
   }

   // verify if edge buffer for backward scan is alocated, if yes desalocate
   if ( m_panBackwardX != NULL ) {
      delete m_panBackwardX;
      m_panBackwardX = NULL;
   }

   // verify if edge buffer for forward scan in texture is alocated
   if ( m_pastForwardTex != NULL ) {
      delete m_pastForwardTex;
      m_pastForwardTex = NULL;
   }

   // verify if edge buffer for backward scan in texture is alocated
   if ( m_pastBackwardTex != NULL ) {
      delete m_pastBackwardTex;
      m_pastBackwardTex = NULL;
   }
}



/* Set all frame buffer with a specific color.
**********************************************************/
void CDraw::mClear ( BYTE yCol )
{
   memset ( (void*) *m_poFrBuff, yCol, m_nSize ); // (void*) *m_poFrBuff is
                                         // one way to access memory buffer.
}



/* Put one pixel on frame buffer.
**********************************************************/
void CDraw::mSetPixel ( INT nX, INT nY, BYTE yCol )
{
   // do clip
   if ( nX >= 0 && nX < m_nResX && nY >= 0 && nY < m_nResY )
      memset ( (void*) m_poFrBuff->m_payMem + nX + ( nY * m_nResX ), yCol, 1 );
}



/* Copy an other frame buffer over the current frame buffer.
oFrBuffSource is the source frame buffer and must have
the same size as the current frame buffer.
**********************************************************/
void CDraw::mCopy ( CMem& oFrBuffSource )
{
   memcpy ( (void*) *m_poFrBuff, (void*) oFrBuffSource, m_poFrBuff->m_dwSize );
}



/* 2 - Copy an other frame buffer (or bitmap) over the
current frame buffer. (nOX, nOY) are the frame buffer
position for the upper left bitmap point (these values
start in upper left frame buffer corner).
nWidth and nHeidht are the dimensions of the bitmap.
No test if bitmap is bigger that the frame buffer in any axis (GPF)!
**********************************************************/
void CDraw::mCopy ( INT nOX, INT nOY, INT nWidth, INT nHeight, CMem& oFrBuffSource )
{
   // copy each bitmap line, no clip!
   for ( INT ncY = 0; ncY < nHeight; ncY ++ )

      memcpy ( (void*) m_poFrBuff->m_payMem + ( ncY * m_nResX ) + ( nOY * m_nResX ) + nOX,
               (void*) oFrBuffSource.m_payMem + ncY * nWidth,
               nWidth );
}



/* Draw a line with the specific color. Low optimized code.
Give the points (nXI, nYI) -> (nXF,nYF). Do line clip.
**********************************************************/
void CDraw::mLine ( INT nXI, INT nYI, INT nXF, INT nYF, BYTE yCol )
{
   // define start and end Y (the less and greater Y)
   INT
      nYStart = nYI,
      nYEnd = nYF,
      nSignalY = 1;
   if ( nYF < nYStart ) {
      nYStart = nYF;
      nYEnd = nYI;
      nSignalY = -1;
   }

   // define start and end X (the less and greater X)
   INT
      nXStart = nXI,
      nXEnd = nXF,
      nSignalX = 1;
   if ( nXF < nXStart ) {
      nXStart = nXF;
      nXEnd = nXI;
      nSignalX = -1;
   }

   // compute delta between the given points
   INT
      nDeltaX = nXF - nXI,
      nDeltaY = nYF - nYI;

   // if delta X is the greater you draw the line using Y as function of X
   if ( abs ( nDeltaX ) > abs ( nDeltaY ) ) {

      float
         fYRate = (float) nDeltaY / nDeltaX, // define Y rate
         fY;                                            // start Y
      // define start line
      if ( nSignalY == -1 )
         fY = (float) nYEnd;
      else
         fY = (float) nYStart;

      // draw line
      // test if X is growing left (1) or right (-1)
      if ( nSignalX == 1 )
         for ( INT nX = nXStart; nX < nXEnd + 1; nX ++ ) {

            // do clip
            if ( nX >= 0 && nX <= m_nMaxX && (INT) fY >= 0 && (INT) fY <= m_nMaxY )
               memset ( (void*) m_poFrBuff->m_payMem + nX + ( (INT) fY * m_nResX ), yCol, 1 );

            fY += fYRate;
         }
      else {

         fYRate = - fYRate;

         for ( INT nX = nXEnd; nX > nXStart; nX -- ) {

            // do clip
            if ( nX >= 0 && nX <= m_nMaxX && (INT) fY >= 0 && (INT) fY <= m_nMaxY )
               memset ( (void*) m_poFrBuff->m_payMem + nX + ( (INT) fY * m_nResX ), yCol, 1 );

            fY += fYRate;
         }
      }
   }
   else {   // draw the line using X as function of Y

      float
         fXRate = (float) nDeltaX / nDeltaY, // define X rate
         fX;                                            // start X
      // define start colum
      if ( nSignalX == -1 )
         fX = (float) nXEnd;
      else
         fX = (float) nXStart;

      // draw line
      // test if Y is growing down (1) or up (-1)
      if ( nSignalY == 1 )
         for ( INT nY = nYStart; nY < nYEnd + 1; nY ++ ) {

            // do clip
            if ( (INT) fX >= 0 && (INT) fX <= m_nMaxX && nY >= 0 && nY <= m_nMaxY )
               memset ( (void*) m_poFrBuff->m_payMem + (INT) fX + ( nY * m_nResX ), yCol, 1 );

            fX += fXRate;
         }
      else {

         fXRate = - fXRate;

         for ( INT nY = nYEnd; nY > nYStart; nY -- ) {

            // do clip
            if ( (INT) fX >= 0 && (INT) fX <= m_nMaxX && nY >= 0 && nY <= m_nMaxY )
               memset ( (void*) m_poFrBuff->m_payMem + (INT) fX + ( nY * m_nResX ), yCol, 1 );

            fX += fXRate;
         }
      }
   }
}



/* Reset the given edge buffer (pastEdgeBuffer): set the
start limit to the minimum x and the end limit to maximum
x acording the frame buffer. Do not perform clipping.
**********************************************************/
void CDraw::mResetClipEdgeBuffer ( tagInterval* pastEdgeBuffer )
{
   tagInterval
      *pstInt = pastEdgeBuffer;

   // scan all edge buffer setting their values
   for ( INT ncYScanLine = 0; ncYScanLine < m_nResY; ncYScanLine ++ ) {
      pstInt->nStart = 0;
      pstInt->nEnd = m_nMaxX;
      pstInt ++;
   }
}



/* Set the given clip buffer using the edges defined by the
given polygon.
The given polygon is clipped through the given edge buffer.
pastVertexes have the array of polygon vertexes in clockwise
or anticlockwise order used to define the edge buffer.
nNumVertexes is the number of polygon vertexes.
pastEdgeBuffer is the clip edge buffer, must have the same size
stClipBox is the bounding box around the clip polygon that can be
used in fast poly clip, enter the current clip box and return the new.
as the vertical resolution (Y).
return TRUE - poly drawn (not totaly clipped) / FALSE - poly
totaly invisible (completely clipped)
**********************************************************/
BOOL CDraw::mSetClipEdgeBuffer ( tagXY* pastVertexes, INT nNumVertexes,
                                 tagInterval* pastEdgeBuffer,
                                 tagRectangle& stClipBox )
{
   // flag if poly has some part (or completely) visible
   BOOL
      bVisible = FALSE;        // originaly not visible
   
   // verify if have enough number of vertexes
   if ( nNumVertexes < 3 )
      ABORT ( MSG_ERROR_LOW_VERTEXES, __FILE__, __LINE__ )

   // seach top and bottom vertexes
   INT
      nYMin = pastVertexes [ 0 ].nY,   // y of the top vertex
      nYMax = pastVertexes [ 0 ].nY,   // y of the bottom vertex
      nPosMin = 0,       // the index of the top vertex (upper in screen)
      nPosMax = 0;       // the index of the bottom vertex (lower in screen)
   // verify vertexes
   for ( INT ncVert = 1; ncVert < nNumVertexes; ncVert ++ ) {

      // verify if array vertex is greater than the current bottom vertex
      if ( pastVertexes [ ncVert ].nY > nYMax ) {
         nYMax = pastVertexes [ ncVert ].nY;
         nPosMax = ncVert;
      }

      // verify if array vertex is less than the current top vertex
      if ( pastVertexes [ ncVert ].nY < nYMin ) {
         nYMin = pastVertexes [ ncVert ].nY;
         nPosMin = ncVert;
      }
   }

   // verify if top and bottom are the same, if yes abort edge fill
   if ( nYMax == nYMin ) {
      stClipBox.nStartY = nYMax;
      stClipBox.nEndY = nYMax;
      return bVisible;
   }

   // scan forward in array of vertexes starting on top and going down
   // With this you scan one side of the polygon.
   INT
      nPos1 = nPosMin,         // the first vertex to be tested
      nPos2 = nPos1;           // the next vertex (used in same test)
   while ( nPos1 != nPosMax ) {

      // define the next vertex position in vertex array
      nPos2 ++;
      // if no more vertexes go to array start
      if ( nPos2 == nNumVertexes )
         nPos2 = 0;

      // define polygon edges, the X for each scanline defined by the line
      // vertex 1 ----> vertex 2
      mScanFlatEdge ( m_panForwardX, pastVertexes [ nPos1 ], pastVertexes [ nPos2 ] );

      // define current vertex as the next vertex
      nPos1 = nPos2;
   }

   // scan backward in array of vertexes starting on top and going down
   // With this you scan one side of the polygon.
   nPos1 = nPosMin;         // the first vertex to be tested
   nPos2 = nPos1;           // the next vertex (used in same test)
   while ( nPos1 != nPosMax ) {

      // define the previous vertex position in vertex array
      nPos2 --;
      // if no more vertexes go to array end
      if ( nPos2 == -1 )
         nPos2 = nNumVertexes - 1;

      // define polygon edges, the X for each scanline defined by the line
      // vertex 1 ----> vertex 2
      mScanFlatEdge ( m_panBackwardX, pastVertexes [ nPos1 ], pastVertexes [ nPos2 ] );

      // define current vertex as the previous vertex
      nPos1 = nPos2;
   }

   // Update edge buffer using 3 loops: the first set the area
   // between clip buffer start and clipping polygon begin. The second
   // set the area over the clipping polygon. The third set the area
   // between end clipping polygon and end buffer.
   // define start buffer position (Y) changed by clipping polygon
   INT
      nY = nYMin;
   if ( nY < 0 )
      nY = 0;     // start buffer line changed by polygon edges
   // define bottom buffer position (Y) changed by clipping polygon
   INT
      nYBottom = nYMax;
   if ( nYBottom > m_nMaxY )
      nYBottom = m_nMaxY;
   // start edge buffer position
   tagInterval
      *pstInt = pastEdgeBuffer,
      *pstEnd = &pastEdgeBuffer [ m_nMaxY ];

   // verify if can change the Y up clip box limit
   if ( nY > stClipBox.nStartY )
      stClipBox.nStartY = nY;
   // verify if can change the Y bottom clip box limit
   if ( nYBottom < stClipBox.nEndY )
      stClipBox.nEndY = nYBottom;

   // store the less and greatter X for this clip edge buffer
   tagInterval
      stXLim;
   // REVIEW
   //stXLim.nStart = m_nMaxX;
   //stXLim.nEnd = 0;

   // scan edge buffer begin to start polygon edges
   if ( nY < m_nResY )
      for ( INT ncY = 0; ncY < nY; ncY ++ ) {
         // set edge buffer: no space available for drawing
         if ( pstInt <= pstEnd ) {
            pstInt->nStart = m_nMaxX;
            pstInt->nEnd = 0;
            if ( pstInt < pstEnd ) 
               pstInt ++;
         }
      }

   // scan edge buffer from top polygon to bottom polygon
   // if X values for this scanline are swapped
   if ( m_panBackwardX [ nY ] < m_panForwardX [ nY ] )
      while ( nY <= nYBottom ) {

         if ( nY < m_nResY ) {

            // make edge buffer X clip and define new buffer X limits
            // clip start X
            if ( * ( m_panBackwardX + nY ) > pstInt->nStart ) {
               if ( pstInt <= pstEnd )
                  if ( * ( m_panBackwardX + nY ) < 0 )
                     pstInt->nStart = 0;
                  else
                     pstInt->nStart = * ( m_panBackwardX + nY );

               // search the minimum X in clip edge buffer
               if ( pstInt->nStart < stXLim.nStart )
                  stXLim.nStart = pstInt->nStart;
            }

            // clip end X
            if ( * ( m_panForwardX + nY ) < pstInt->nEnd ) {
               if ( pstInt <= pstEnd ) 
                  if ( * ( m_panForwardX + nY ) > m_nMaxX )
                     pstInt->nEnd = m_nMaxX;
                  else
                     pstInt->nEnd = * ( m_panForwardX + nY );
   
               // search the maximum X in clip edge buffer
               if ( pstInt->nEnd > stXLim.nEnd )
                  stXLim.nEnd = pstInt->nEnd;
            }

            // if line visible
            if ( pstInt->nStart >= stClipBox.nStartX && pstInt->nEnd <= stClipBox.nEndX )
               bVisible = TRUE;   // poly is visible (partialy or completely)

            if ( pstInt < pstEnd ) 
               pstInt ++;
         }
         
         // goto next scanline
         nY ++;
      }
   else 
      while ( nY <= nYBottom ) {

         if ( nY < m_nResY ) {

            // make edge buffer X clip and define new buffer X limits
            // clip start X
            if ( * ( m_panForwardX + nY ) > pstInt->nStart ) {
               if ( pstInt <= pstEnd ) 
                  if ( * ( m_panForwardX + nY ) < 0 )
                     pstInt->nStart = 0;
                  else
                     pstInt->nStart = * ( m_panForwardX + nY );

               // search the minimum X in clip edge buffer
               if ( pstInt->nStart < stXLim.nStart )
                  stXLim.nStart = pstInt->nStart;
            }

            // clip end X
            if ( * ( m_panBackwardX + nY ) < pstInt->nEnd ) {
               if ( pstInt <= pstEnd )
                  if ( * ( m_panBackwardX + nY ) > m_nMaxX )
                     pstInt->nEnd = m_nMaxX;
                  else
                     pstInt->nEnd = * ( m_panBackwardX + nY );

               // search the maximum X in clip edge buffer
               if ( pstInt->nEnd > stXLim.nEnd )
                  stXLim.nEnd = pstInt->nEnd;
            }

            // if line visible
            if ( pstInt->nStart >= stClipBox.nStartX && pstInt->nEnd <= stClipBox.nEndX )
               bVisible = TRUE;   // poly is visible (partialy or completely)

            if ( pstInt < pstEnd ) 
               pstInt ++;
         }
         
         // goto next scanline
         nY ++;
      }

   // scan end clip polygon to buffer end
   for ( INT ncY = nYBottom + 1; ncY < m_nResY; ncY ++ ) {
      if ( nY < m_nResY ) {
         // set edge buffer: no space available for drawing
         if ( pstInt <= pstEnd ) {
            pstInt->nStart = m_nMaxX;
            pstInt->nEnd = 0;
         }
         if ( pstInt < pstEnd ) 
            pstInt ++;
      }
   }

   // REVIEW
   // verify if have updated X start and end, if no use original values
   //if ( bVisible ) {  // poly visible, must have good values for X limits
      //if ( stXLim.nStart == stClipBox.nEndX )
         //stXLim.nStart = stClipBox.nStartX;
      //if ( stXLim.nEnd == stClipBox.nStartX )
         //stXLim.nEnd = stClipBox.nEndX;
   //}

   // verify if can update clip box X limits
   /*if ( stClipBox.nStartX < stXLim.nStart )
      stClipBox.nStartX = stXLim.nStart;
   if ( stClipBox.nEndX > stXLim.nEnd )
      stClipBox.nEndX = stXLim.nEnd;*/

/*for (int n=0;n<320;n++){
   if(pastEdgeBuffer[n].nStart<0)
      ABORT ( "baixo", __FILE__, __LINE__ )
   if(pastEdgeBuffer[n].nEnd>320)
      ABORT ( "alto", __FILE__, __LINE__ )
}  */


   return bVisible;
}



/* Draw a flat polygon with only 4 vertex in frame buffer.
Medium optimized. The vertexes must be in clockwise or
anticlockwise. yCol is the polygon fill color.
This function use in implementation fix point 16.16.
Perform clip.

                        (x4, y4)
                        *                -
                       / \                 loop
                      /   \                    |
            (x1, y1) *     \                   |
                      \     * (x3, y3)         | compute x positions
                       \   /                   | between edges
                        \ /                    V
                         *               -
                         (x2, y2)

Line equation:
Y = aX + b, where
            a is a increase rate
            b is the offset of the line in y axis
Equation used in intersections (how x change for each scanline + x vertex)
X = ( deltaY * increaseX ) + Xvertex, where
            increaseX is 1 / a, how x change per scanline
            deltaY is the distance between a Y vertex and Y loop
            Xvertex is the X in vertex used in computation of deltaY

HORIZONTAL LINE DRAW (draw 4 bytes in same time, use strings instructions):
__asm__ __volatile__ ("

      movl  %0,      %%edi    destin
      movl  %1,      %%ecx    number of interactions (line length)
      movb  %2,      %%eax    color (repeat color 4 times - 32 bits)
                              If length is not multiple of 4 draw line by
                              4 in 4 bytes and the rest. Start drawing rest.
      movb  %%cl,    %%bl     load bh with last 8 bits to test
      andb  $3,      %%bl     select only 2 last bits (the rest)
      cmpb  $3,      %%bl     verify if rest is 3, if yes draw 3 pixels
      je    Draw3             goto draw continuous
      cmpb  $2,      %%bl     verify if rest is 2, if yes draw 2 pixels
      je    Draw2             goto draw continuous
      cmpb  $1,      %%bl     verify if rest is 1, if yes draw 1 pixels
      je    Draw1             goto draw continuous
      jmp   DrawAll           if no goto executed the rest is 0
   Draw3:                     if rest 3
      stosb
   Draw2:                     if rest 2 or 3
      stosb
   Draw1:                     if rest 3 or 2 or 1
      stosb

   DrawAll:                   draw line (do not test rest)
      shrl  $2,     %%ecx     remove interactions done in rest cases (/4)
      cld                     clear direction flag
      rep                     init repet until ecx equal 0 (end of line)
         stosl                put 4 pixels (the value in eax)
      "
      :
      : "g" ( uEndLine + nXIni ), "g" ( nXEnd - nXIni + 1 ), "g" ( uCol )
      : "%bl", "%ecx", "%edi", "eax"
);
**********************************************************/
void CDraw::mFlatPolygon4 ( INT nX1, INT nY1, INT nX2, INT nY2,
                            INT nX3, INT nY3, INT nX4, INT nY4, BYTE yCol )
{
   // clip y edges
   // define less (top) and greater (bottom) y values
   INT
      nYSup = nY1,
      nYInf = nY1;
   // test y2
   if ( nYSup > nY2 )
      nYSup = nY2;
   else
      if ( nYInf < nY2 )
         nYInf = nY2;
   // test y3
   if ( nYSup > nY3 )
      nYSup = nY3;
   else
      if ( nYInf < nY3 )
         nYInf = nY3;
   // test y4
   if ( nYSup > nY4 )
      nYSup = nY4;
   else
      if ( nYInf < nY4 )
         nYInf = nY4;
   // test if points are outside frame buffer limits, if yes change values
   if ( nYSup < 0 )
      nYSup = 0;
   if ( nYInf > m_nMaxY )
      nYInf = m_nMaxY;

   // define if be possible compute x increase rates, if yes do
   INT
      nIncX12 = 0,
      nIncX14 = 0,
      nIncX23 = 0,
      nIncX34 = 0;
   BOOL
      bIncValid12 = TRUE,
      bIncValid14 = TRUE,
      bIncValid23 = TRUE,
      bIncValid34 = TRUE;
   // be computed valores only for points that are not in same lines
   if ( nY1 != nY2 )
      nIncX12 = (INT) ( INCREASEX ( nX1, nY1, nX2, nY2 ) * 65536 );
   else
      bIncValid12 = FALSE;
   if ( nY1 != nY4 )
      nIncX14 = (INT) ( INCREASEX ( nX1, nY1, nX4, nY4 ) * 65536 );
   else
      bIncValid14 = FALSE;
   if ( nY2 != nY3 )
      nIncX23 = (INT) ( INCREASEX ( nX2, nY2, nX3, nY3 ) * 65536 );
   else
      bIncValid23 = FALSE;
   if ( nY3 != nY4 )
      nIncX34 = (INT) ( INCREASEX ( nX3, nY3, nX4, nY4 ) * 65536 );
   else
      bIncValid34 = FALSE;

   // define address of the beging for the first line in frame buffer
   UINT
      pLineAddress = (PTRVAL) m_poFrBuff->m_payMem + m_nResX * nYSup;

   // define color fill
   UINT
      uCol = yCol | ( yCol << 8 ) | ( yCol << 16 ) | ( yCol << 24 );

   // goto polygon start to end in Y axis
   INT
      nXIni,    // start x in scanline
      nXEnd,    // end x in scanline
      nX;       // x value of intersection by 2 lines
   for ( INT ncY = nYSup; ncY <= nYInf; ncY ++ ) {

      // initial x values in scanline
      nXIni = 32767;
      nXEnd = - 32768;

      // compute x between vertex 1 and 4
      if ( bIncValid14 &&
           ( ncY >= nY1 && ncY <= nY4 ) ||
           ( ncY <= nY1 && ncY >= nY4 ) ) {

         nX = (INT) ( ( ncY - nY4 ) * nIncX14 >> 16 ) + nX4;

         // verify if start x must be changed
         if ( nX < nXIni )
            nXIni = nX;
         // verify if end x must be changed
         if ( nX > nXEnd )
            nXEnd = nX;
      }

      // compute x between vertex 1 and 2
      if ( bIncValid12 &&
           ( ncY >= nY1 && ncY <= nY2 ) ||
           ( ncY <= nY1 && ncY >= nY2 ) ) {

         nX = (INT) ( ( ncY - nY2 ) * nIncX12 >> 16 ) + nX2;

         // verify if start x must be changed
         if ( nX < nXIni )
            nXIni = nX;
         // verify if end x must be changed
         if ( nX > nXEnd )
            nXEnd = nX;
      }

      // compute x between vertex 2 and 3
      if ( bIncValid23 &&
           ( ncY >= nY2 && ncY <= nY3 ) ||
           ( ncY <= nY2 && ncY >= nY3 ) ){

         nX = (INT) ( ( ncY - nY3 ) * nIncX23 >> 16 ) + nX3;

         // verify if start x must be changed
         if ( nX < nXIni )
            nXIni = nX;
         // verify if end x must be changed
         if ( nX > nXEnd )
            nXEnd = nX;
      }

      // compute x between vertex 3 and 4
      if ( bIncValid34 &&
           ( ncY >= nY3 && ncY <= nY4 ) ||
           ( ncY <= nY3 && ncY >= nY4 ) ) {

         nX = (INT) ( ( ncY - nY4 ) * nIncX34 >> 16 ) + nX4;

         // verify if start x must be changed
         if ( nX < nXIni )
            nXIni = nX;
         // verify if end x must be changed
         if ( nX > nXEnd )
            nXEnd = nX;
      }

      // clip x values
      // verify start
      if ( nXIni < 0 )
         nXIni = 0;
      else
         // start can be in right (line outside frame buffer)
         if ( nXIni > m_nMaxX ) {
            // compute next scanline address
            pLineAddress += m_nResX;
            // do not draw line
            continue;
         }
      // verify end
      if ( nXEnd > m_nMaxX )
         nXEnd = m_nMaxX;
      else
         // end can be in left (line outside frame buffer)
         if ( nXEnd < 0 ) {
            // compute next scanline address
            pLineAddress += m_nResX;
            // do not draw line
            continue;
         }

      // draw horizontal line
      // if compile with asm code
      #ifdef COMP_ASM
      __asm__ __volatile__ ("

            movl  %0,      %%edi
            movl  %1,      %%ecx
            movl  %2,      %%eax

            movb  %%cl,    %%bl
            andb  $3,      %%bl
            cmpb  $3,      %%bl
            je    DrawPix3
            cmpb  $2,      %%bl
            je    DrawPix2
            cmpb  $1,      %%bl
            je    DrawPix1
            jmp   DrawPixAll
         DrawPix3:
            stosb
         DrawPix2:
            stosb
         DrawPix1:
            stosb

         DrawPixAll:
            shrl  $2,      %%ecx
            cld
            rep
               stosl
            "
            :
            : "g" ( pLineAddress + nXIni ), "g" ( nXEnd - nXIni + 1 ), "g" ( uCol )
            : "%bl", "cl", "%ecx", "%edi", "eax"
      );
      #endif
      // compile C only (no asm code)
      #ifndef COMP_ASM
      memset ( (void*) ( pLineAddress + nXIni ), uCol, nXEnd - nXIni + 1 );
      #endif

      // compute next scanline address
      pLineAddress += m_nResX;
   }
}



/* Draw a flat polygon with any number of vertexes (points).
pastVertexes have the array of polygon vertexes in
clockwise or anticlockwise order. nNumVertexes is the
number of vertexes in array. yCol is the polygon fill color.
Perform clip. Medium optimized.
**********************************************************/
void CDraw::mFlatPolygon ( tagXY* pastVertexes,
                           INT nNumVertexes, BYTE yCol )
{
   // verify if have enough number of vertexes
   if ( nNumVertexes < 3 )
      ABORT ( MSG_ERROR_LOW_VERTEXES, __FILE__, __LINE__ )

   // seach top and bottom vertexes
   INT
      nYMin = pastVertexes [ 0 ].nY,   // y of the top vertex
      nYMax = pastVertexes [ 0 ].nY,   // y of the bottom vertex
      nPosMin = 0,       // the index of the top vertex (upper in screen)
      nPosMax = 0;       // the index of the bottom vertex (lower in screen)
   // verify vertexes
   for ( INT ncVert = 1; ncVert < nNumVertexes; ncVert ++ ) {

      // verify if array vertex is greater than the current bottom vertex
      if ( pastVertexes [ ncVert ].nY > nYMax ) {
         nYMax = pastVertexes [ ncVert ].nY;
         nPosMax = ncVert;
      }

      // verify if array vertex is less than the current top vertex
      if ( pastVertexes [ ncVert ].nY < nYMin ) {
         nYMin = pastVertexes [ ncVert ].nY;
         nPosMin = ncVert;
      }
   }

   // verify if top and bottom are the same, if yes abort polygon fill
   if ( nYMax == nYMin )
      return;

   // define color fill
   UINT
      uCol = yCol | ( yCol << 8 ) | ( yCol << 16 ) | ( yCol << 24 );

   // scan forward in array of vertexes starting on top and going down
   // With this you scan one side of the polygon.
   INT
      nPos1 = nPosMin,         // the first vertex to be tested
      nPos2 = nPos1;           // the next vertex (used in same test)
   while ( nPos1 != nPosMax ) {

      // define the next vertex position in vertex array
      nPos2 ++;
      // if no more vertexes go to array start
      if ( nPos2 == nNumVertexes )
         nPos2 = 0;

      // define polygon edges, the X for each scanline defined by the line
      // vertex 1 ----> vertex 2
      mScanFlatEdge ( m_panForwardX, pastVertexes [ nPos1 ], pastVertexes [ nPos2 ] );

      // define current vertex as the next vertex
      nPos1 = nPos2;
   }

   // scan backward in array of vertexes starting on top and going down
   // With this you scan one side of the polygon.
   nPos1 = nPosMin;         // the first vertex to be tested
   nPos2 = nPos1;           // the next vertex (used in same test)
   while ( nPos1 != nPosMax ) {

      // define the previous vertex position in vertex array
      nPos2 --;
      // if no more vertexes go to array end
      if ( nPos2 == -1 )
         nPos2 = nNumVertexes - 1;

      // define polygon edges, the X for each scanline defined by the line
      // vertex 1 ----> vertex 2
      mScanFlatEdge ( m_panBackwardX, pastVertexes [ nPos1 ], pastVertexes [ nPos2 ] );

      // define current vertex as the previous vertex
      nPos1 = nPos2;
   }

   // draw polygon
   // define scanline Y (if nYMin is less than 0, outside frame buff, use 0)
   INT
      nY = nYMin < 0 ? 0 : nYMin;
   // define memory position
   PTRVAL
      pLineAddress = (PTRVAL) m_poFrBuff->m_payMem + nY * m_nResX;

   // if X values for this scanline are swapped
   if ( m_panBackwardX [ nY ] < m_panForwardX [ nY ] )

      // while Y is upper than bottom vertex or bottom frame buffer
      while ( nY < nYMax && nY < m_nResY ) {
   
         mHorizontalLine ( m_panBackwardX [ nY ], m_panForwardX [ nY ], uCol, pLineAddress );

         // goto next scanline
         nY ++;
         pLineAddress += m_nResX;
      }
   else
      // while Y is upper than bottom vertex or bottom frame buffer
      while ( nY < nYMax && nY < m_nResY ) {
         
         mHorizontalLine ( m_panForwardX [ nY ], m_panBackwardX [ nY ], uCol, pLineAddress );

         // goto next scanline
         nY ++;
         pLineAddress += m_nResX;
      }
}



/* Draw a affine texture mapped polygon on frame buffer.
Make clip through the given edge buffer.
The polygon must have vertexes in clockwise or anticlockwise.
pastPolVertexes is the array of polygon vertexes;
pastTexVertexes is the array of texture vertexes;
One vertex in polygon has one vertex in texture.
All vertexes must be in clockwise or anticlockwise order.
nNumVertex is the number of polygon vertexes.
oTx is the texture to use in polygon.
fNormLightLevel has the normalized polygon light level (between 0 and 1)
pastEdgeBuffer is the clip buffer to be used,
if set to NULL only frame buffer clip is made (default NULL).
**********************************************************/
void CDraw::mTexturedPolygon ( tagXY* pastPolVertexes,
                               tagXY* pastTexVertexes,
                               INT nNumVertexes, CTexture& oTx,
                               tagInterval* pastEdgeBuffer,
                               float fNormLightLevel )
{
   // verify if have enough number of vertexes
   if ( nNumVertexes < 3 )
      ABORT ( MSG_ERROR_LOW_VERTEXES, __FILE__, __LINE__ )

   // seach top and bottom vertexes
   INT
      nYMin = pastPolVertexes [ 0 ].nY,   // y of the top vertex
      nYMax = pastPolVertexes [ 0 ].nY,   // y of the bottom vertex
      nPosMin = 0,       // the index of the top vertex (upper in screen)
      nPosMax = 0;       // the index of the bottom vertex (lower in screen)
   // verify vertexes
   for ( INT ncVert = 1; ncVert < nNumVertexes; ncVert ++ ) {

      // verify if array vertex is greater than the current bottom vertex
      if ( pastPolVertexes [ ncVert ].nY > nYMax ) {
         nYMax = pastPolVertexes [ ncVert ].nY;
         nPosMax = ncVert;
      }

      // verify if array vertex is less than the current top vertex
      if ( pastPolVertexes [ ncVert ].nY < nYMin ) {
         nYMin = pastPolVertexes [ ncVert ].nY;
         nPosMin = ncVert;
      }
   }

   // verify if top and bottom are the same, if yes abort polygon fill
   if ( nYMax == nYMin )
      return;

   // scan forward in array of vertexes starting on top and going down
   // With this you scan one side of the polygon.
   INT
      nPos1 = nPosMin,         // the first vertex to be tested
      nPos2 = nPos1;           // the next vertex (used in same test)
   while ( nPos1 != nPosMax ) {

      // define the next vertex position in vertex array
      nPos2 ++;
      // if no more vertexes go to array start
      if ( nPos2 == nNumVertexes )
         nPos2 = 0;

      // define polygon and texture edges, the X for each scanline defined
      // by the line vertex 1 ----> vertex 2
      mScanTexturedEdge ( m_panForwardX, m_pastForwardTex,
                          pastPolVertexes [ nPos1 ],
                          pastPolVertexes [ nPos2 ],
                          pastTexVertexes [ nPos1 ],
                          pastTexVertexes [ nPos2 ] );

      // define current vertex as the next vertex
      nPos1 = nPos2;
   }

   // scan backward in array of vertexes starting on top and going down
   // With this you scan one side of the polygon.
   nPos1 = nPosMin;         // the first vertex to be tested
   nPos2 = nPos1;           // the next vertex (used in same test)
   while ( nPos1 != nPosMax ) {

      // define the previous vertex position in vertex array
      nPos2 --;
      // if no more vertexes go to array end
      if ( nPos2 == -1 )
         nPos2 = nNumVertexes - 1;

      // define polygon and texture edges, the X for each scanline defined
      // by the line vertex 1 ----> vertex 2
      mScanTexturedEdge ( m_panBackwardX, m_pastBackwardTex,
                          pastPolVertexes [ nPos1 ],
                          pastPolVertexes [ nPos2 ],
                          pastTexVertexes [ nPos1 ],
                          pastTexVertexes [ nPos2 ] );

      // define current vertex as the previous vertex
      nPos1 = nPos2;
   }

   // define the light level for the loaded light table
   // transform normalized light level to light table value
   DWORD
      dwLightLevel = (DWORD) ( fNormLightLevel * ( m_poLightTable->m_dwNumLevels - 1 ) );

   // draw polygon
   // define scanline Y (if nYMin is less than 0, outside frame buff, use 0)
   INT
      nY = ( nYMin < 0 ) ? 0 : nYMin;
   // define memory position
   PTRVAL
      pLineAddress = (PTRVAL) m_poFrBuff->m_payMem + nY * m_nResX;

   // verify if clipping edge buffer exist, if yes use else use default
   tagInterval
      stDefaultClip;
   BYTE
      yNextLine = 1; // to go to next clip buffer element when scanning lines
   if ( pastEdgeBuffer == NULL ) {
      // define the default clip
      stDefaultClip.nStart = 0;
      stDefaultClip.nEnd = m_nMaxX;
      pastEdgeBuffer = &stDefaultClip;
      // no line change
      yNextLine = 0;
   }
   else
      // find the right position to start buffer scan
      pastEdgeBuffer += nY;

   // if X values for this scanline are swapped
   if ( m_panBackwardX [ nY ] < m_panForwardX [ nY ] )
      // while Y is upper than bottom vertex or bottom frame buffer
      while ( nY < nYMax && nY < m_nResY ) {

         mHorizontalTexturedLine ( m_panBackwardX [ nY ],
                                   m_panForwardX [ nY ],
                                   m_pastBackwardTex [ nY ],
                                   m_pastForwardTex [ nY ],
                                   *pastEdgeBuffer,
                                   (BYTE*) oTx, pLineAddress,
                                   dwLightLevel);
         // goto next scanline
         nY ++;
         pLineAddress += m_nResX;
         pastEdgeBuffer += yNextLine;
      }
   else
      // while Y is upper than bottom vertex or bottom frame buffer
      while ( nY < nYMax && nY < m_nResY ) {

         mHorizontalTexturedLine ( m_panForwardX [ nY ],
                                   m_panBackwardX [ nY ],
                                   m_pastForwardTex [ nY ],
                                   m_pastBackwardTex [ nY ],
                                   *pastEdgeBuffer,
                                   (BYTE*) oTx, pLineAddress,
                                   dwLightLevel);

         // goto next scanline
         nY ++;
         pLineAddress += m_nResX;
         pastEdgeBuffer += yNextLine;
      }
}



/* Used by CDrawClip::mSetClipEdgeBuffer to define X limits
for clip edge buffer and by CDraw::mFlatPolygon and to define
the X coordenates in each scanline for the polygon edges.
panEdgeX - the x coordenates for each scanline;
stV1 and stV2 - vertex between scan must be done.
Always stV1 is less (topper) than stV2 in Y axis.
**********************************************************/
void CDraw::mScanFlatEdge ( INT* panEdgeX, tagXY& stV1,
                                           tagXY& stV2 )
{
   // verify if Y are in same line, if yes exit
   if ( stV1.nY == stV2.nY )
      return;

   INT
      // define where X start in first scanline, use fix point 16.16
      nX = stV1.nX << 16,
      // X increase rate
      nXInc = ( ( stV2.nX - stV1.nX ) << 16 ) /
                ( stV2.nY - stV1.nY );

   // do Y clip in frame buffer limits
   INT
      nYTop,
      nYBottom;
   // if y for vertex 1 is upper than frame buffer limits
   if ( stV1.nY < 0 ) {
      nYTop = 0;
      // if top Y has been cliped then is necessary update initial X value
      nX -= stV1.nY * nXInc;
   }
   else
      nYTop = stV1.nY;
   // if y for vertex 2 is lower than frame buffer limits
   if ( stV2.nY > m_nMaxY )
      nYBottom = m_nMaxY;
   else
      nYBottom = stV2.nY;

   // verify if need clip
   if ( nYBottom <= nYTop )
      return;

   // define X values in each scanline between top vertex and bottom vertex
   INT
      nYScan = nYTop;    // scanline in test
   do {

      panEdgeX [ nYScan ] = nX >> 16;    // X in frame buff, convert fix->int
      nX += nXInc;                       // new X in next scanline
      nYScan ++;                         // next scanline

   } while ( nYScan <= nYBottom );       // while not found bottom vertex
}



/* Used by CDraw::mTexturedPolygon to define the X coordenates
in each scanline for the polygon and texture edges.
panEdgePol - the x coordenates for each scanline in polygon;
panEdgeTex - the x and y coordenates for each scanline in texture;
stP1 and stP2 - polygon vertex between scan must be done.
stT1 and stT2 - texture vertex between scan must be done.
Always stP1 is less (topper) than stP2 in Y axis.
Use fix point 16.16.
**********************************************************/
void CDraw::mScanTexturedEdge ( INT* panEdgePol, tagXY* pastEdgeTex,
                                tagXY& stP1, tagXY& stP2,
                                tagXY& stT1, tagXY& stT2 )
{
   // verify if Y are in same line, if yes exit
   if ( stP1.nY == stP2.nY )
      return;

   INT
      // define where X start in first scanline
      nX = stP1.nX << 16,
      // define where texture X and Y start in first scanline
      nTexX = stT1.nX << 16,
      nTexY = stT1.nY << 16,
      // define X increase rate for each scanline
      nXInc = ( ( stP2.nX - stP1.nX ) << 16 ) / ( stP2.nY - stP1.nY ),
      // define texture X and Y increase rates for each scan line
      nTexXInc = ( ( stT2.nX - stT1.nX ) << 16 ) / ( stP2.nY - stP1.nY ),
      nTexYInc = ( ( stT2.nY - stT1.nY ) << 16 ) / ( stP2.nY - stP1.nY );

   // do Y clip in frame buffer limits
   INT
      nYTop,
      nYBottom;
   // if y for vertex 1 is upper than frame buffer limits
   if ( stP1.nY < 0 ) {
      nYTop = 0;
      // top Y has been cliped then is necessary update values
      nX    -= stP1.nY * nXInc;
      nTexX -= stP1.nY * nTexXInc;
      nTexY -= stP1.nY * nTexYInc;
   }
   else
      nYTop = stP1.nY;
   // if y for vertex 2 is lower than frame buffer limits
   if ( stP2.nY > m_nMaxY )
      nYBottom = m_nMaxY;
   else
      nYBottom = stP2.nY;

   // verify if need clip
   if ( nYBottom <= nYTop )
      return;

   // define X values in each scanline between top vertex and bottom vertex
   INT
      nYScan = nYTop;    // scanline in test

   do {

      panEdgePol [ nYScan ] = nX >> 16;  // X in frame buff, convert fix->int
      ( pastEdgeTex + nYScan )->nX = nTexX; // X in texture
      ( pastEdgeTex + nYScan )->nY = nTexY; // Y in texture
      nX += nXInc;                       // new X in next scanline
      nTexX += nTexXInc;                 // new texture X
      nTexY += nTexYInc;                 // new texture Y
      nYScan ++;                         // next scanline

   } while ( nYScan <= nYBottom );       // while not found bottom vertex
}



/* Draw a horizontal line.
uCol - is an array with 4 bytes. Each elements must have the
color. Example: to draw a line with color 10 create the array:
char chCol [] = { 10, 10, 10, 10 };   pLineAddress is the
position in frame buffer where the line must be drawn.
High optimized. Clip only X values.
**********************************************************/
void CDraw::mHorizontalLine ( INT nXIni, INT nXEnd, UINT uCol,
                              PTRVAL pLineAddress )
{
   // clip x values
   // verify start
   if ( nXIni < 0 )
      nXIni = 0;
   else
      // start can be in right (line outside frame buffer)
      if ( nXIni > m_nMaxX )
         return;
   // verify end
   if ( nXEnd > m_nMaxX )
      nXEnd = m_nMaxX;
   else
      // end can be in left (line outside frame buffer)
      if ( nXEnd < 0 )
         return;

   // draw horizontal line (See coments in CDraw::mFlatPolygon4)
   // if compile with asm code
   #ifdef COMP_ASM
   __asm__ __volatile__ ("

         movl  %0,      %%edi
         movl  %1,      %%ecx
         movl  %2,      %%eax

         movb  %%cl,    %%bl
         andb  $3,      %%bl
         cmpb  $3,      %%bl
         je    Draw3
         cmpb  $2,      %%bl
         je    Draw2
         cmpb  $1,      %%bl
         je    Draw1
         jmp   DrawAll
      Draw3:
         stosb
      Draw2:
         stosb
      Draw1:
         stosb

      DrawAll:
         shrl  $2,      %%ecx
         cld
         rep
            stosl
         "
         :
         : "g" ( pLineAddress + nXIni ), "g" ( nXEnd - nXIni + 1 ), "g" ( uCol )
         : "%bl", "cl", "%ecx", "%edi", "eax"
   );
   #endif
   // compile C only (no asm code)
   #ifndef COMP_ASM
   memset ( (void*) ( pLineAddress + nXIni ), uCol, nXEnd - nXIni + 1 );
   #endif
}



/* Draw a horizontal textured line.
Clip X values through given clip edge X limits.
nXIni - left X value;
nXEnd - right X value;
stTexIni - texture coordenate start;
stTexEnd - texture coordenate end;
stLimits - the X limits to clip the line
pyTex - texture; pLineAddress - where draw line in frame buffer.
pLineAddress is the position in frame buffer where the
line must be drawn.
**********************************************************/
void CDraw::mHorizontalTexturedLine ( INT nXIni, INT nXEnd,
                                      tagXY& stTexIni,
                                      tagXY& stTexEnd,
                                      tagInterval& stLimits,
                                      BYTE* pyTex, PTRVAL pLineAddress,
                                      DWORD dwLightLevel )
{
   // clip x values
   INT
      nFrXIni = nXIni,   // define left X value in frame buffer
      nFrXEnd = nXEnd;   // define right X value in frame buffer
   // verify start
   if ( nXIni < stLimits.nStart )
      nFrXIni = stLimits.nStart;
   // verify end
   if ( nXEnd > stLimits.nEnd )
      nFrXEnd = stLimits.nEnd;

   // verify if line has no lenght
   if ( nFrXEnd <= nFrXIni )
      return;

   INT
      // define where texture X and Y start
      nTexX = stTexIni.nX,
      nTexY = stTexIni.nY,
      // texture X and Y increase rates
      nTexXInc = ( stTexEnd.nX - stTexIni.nX ) / ( nXEnd - nXIni ),
      nTexYInc = ( stTexEnd.nY - stTexIni.nY ) / ( nXEnd - nXIni );

   // if X has been clipped
   if ( nXIni < stLimits.nStart ) {
      nTexX -= ( nXIni - stLimits.nStart ) * nTexXInc;
      nTexY -= ( nXIni - stLimits.nStart ) * nTexYInc;
   }

   // define start and end address in frame buffer
   BYTE
      *pyCurr = (BYTE*) pLineAddress + nFrXIni,   // current address
      *pyEnd = (BYTE*) pLineAddress + nFrXEnd;    // end address

   // line draw loop
   do {
         // compile making lighting
         #ifdef COMP_LIGHTS
      //if ( pyCurr >= m_poFrBuff->m_payMem && pyCurr <= ( m_poFrBuff->m_payMem + m_nResX * m_nResY - 1 ) ) {
         // light it
         *pyCurr = * ( m_poLightTable->m_payLightTable +
                       pyTex [ ( ( nTexX >> LINE_SHIFT_X ) & LINE_AND_X ) + ( ( nTexY >> LINE_SHIFT_Y ) & LINE_AND_Y ) ] *
                       m_poLightTable->m_dwNumLevels + dwLightLevel );
         #endif
         #ifndef COMP_LIGHTS
         // light it
         *pyCurr = pyTex [ ( ( nTexX >> LINE_SHIFT_X ) & LINE_AND_X ) + ( ( nTexY >> LINE_SHIFT_Y ) & LINE_AND_Y ) ];
         #endif

         // update texture coordenates
         nTexX += nTexXInc;
         nTexY += nTexYInc;
      //}
      //else
         //ABORT ("Escreveu fora do frame buffer!\n",__FILE__,__LINE__)
      // go to next pixel
      pyCurr ++;

   } while ( pyCurr <= pyEnd );  // while not found end of line
}
