/* All math functions.

TIPS:
- this object need to huch memory and time to compute the 
  internal lockup tables of sen/cos. Don't alocate the tables if
  you don't need and do not declare one object instance of this
  class inside a inner loop.
*/



#include <string.h>
#include <math.h>

#include "defs.h"

#include "node.h"

#include "nodever.h"

#include "list.h"

#include "polygon.h"

#include "nodepol.h"

#include "math.h"



/* constructor - init math functions. If bBuildTrigTables
is TRUE build the sen/cos tables, else no (default FALSE)
**********************************************************/
CMath::CMath ( BOOL bBuildTrigTables )
{
   m_panSin = m_panCos = NULL;
   m_pafSin = m_pafCos = NULL;

   if ( ! mCreate ( bBuildTrigTables ) )
      ABORT ( MSG_ERROR_MEMORY_LOW, __FILE__, __LINE__ )

}



/* destructor - empty object, if need
**********************************************************/
CMath::~CMath ()
{
   mDestroy ();
}



/* Init math functions. If bBuildTrigTables
is TRUE build the sen/cos tables, else no (default FALSE).
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CMath::mCreate ( BOOL bBuildTrigTables )
{
   mDestroy ();
   
   // make the sen/cos tables
   if ( bBuildTrigTables ) {

      // verify if need alocate the lockup tables
      if ( m_panSin == NULL ) {
         m_panSin = new INT [ ROUND ];
         if ( m_panSin == NULL )
            return FALSE;
      }
      if ( m_panCos == NULL ) {
         m_panCos = new INT [ ROUND ];
         if ( m_panCos == NULL )
            return FALSE;
      }
      if ( m_pafSin == NULL ) {
         m_pafSin = new float [ ROUND ];
         if ( m_pafSin == NULL )
            return FALSE;
      }
      if ( m_pafCos == NULL ) {
         m_pafCos = new float [ ROUND ];
         if ( m_pafCos == NULL )
            return FALSE;
      }

      // sin and cos lockup tables
      for ( INT ncD = 0; ncD < ROUND; ncD ++ )  {
         // fix point 16.16
         m_panSin [ ncD ] = (INT) ( sin ( ncD * PI / ROUND * 2 ) * 65536 );
         m_panCos [ ncD ] = (INT) ( cos ( ncD * PI / ROUND * 2 ) * 65536 );
         // floating point
         m_pafSin [ ncD ] = sin ( ncD * PI / ROUND * 2 );
         m_pafCos [ ncD ] = cos ( ncD * PI / ROUND * 2 );
      }
   }

   return TRUE;
}



/* Desalocate lockup tables, if alocated
**********************************************************/
void CMath::mDestroy ()
{
   if ( m_panSin != NULL ) {
      delete m_panSin;
      m_panSin = NULL;
   }
   if ( m_panCos != NULL ) {
      delete m_panCos;
      m_panCos = NULL;
   }
   if ( m_pafSin != NULL ) {
      delete m_pafSin;
      m_pafSin = NULL;
   }
   if ( m_pafCos != NULL ) {
      delete m_pafCos;
      m_pafCos = NULL;
   }
}



/* Compute the poly plan equation:
First compute polygon normal vector and then the D component.
 Plan equation: AX + BY + CZ + D = 0,   where
 (A, B, C) is a plan normal and unitary vector
 D = AX' + BY' + CZ',   (X', Y', Z') is a point inside plan
 (X, Y, Z) is a point to test
oVertexes - is a list of x,y,z vertexes
poPol - point the poly inside the plan (change it's plan data)
**********************************************************/
void CMath::mDefinePolyPlanEquation ( CList& oVertexes, CNodePolygon* poPol )
{
   // verify if enought vertexes are defined in mesh
   if ( poPol->m_dwNumVertexes < 3 )
      ABORT ( MSG_ERROR_LOW_VERTEXES, __FILE__, __LINE__ )

   // define 3 vertexes inside polygon plan
   CNodeVertex 
      *poVert0 = (CNodeVertex*) oVertexes.mRetNode ( poPol->m_padwVertexes [ 0 ] ),
      *poVert1 = (CNodeVertex*) oVertexes.mRetNode ( poPol->m_padwVertexes [ 1 ] ),
      *poVert2 = (CNodeVertex*) oVertexes.mRetNode ( poPol->m_padwVertexes [ 2 ] );

   // define 2 vectors inside the poly plan
   tagXYZ
      stVector1,
      stVector2;
   stVector1.nX = poVert0->m_stVertex.nX - poVert1->m_stVertex.nX;
   stVector1.nY = poVert0->m_stVertex.nY - poVert1->m_stVertex.nY;
   stVector1.nZ = poVert0->m_stVertex.nZ - poVert1->m_stVertex.nZ;
   stVector2.nX = poVert0->m_stVertex.nX - poVert2->m_stVertex.nX;
   stVector2.nY = poVert0->m_stVertex.nY - poVert2->m_stVertex.nY;
   stVector2.nZ = poVert0->m_stVertex.nZ - poVert2->m_stVertex.nZ;

   // compute normal vector
   tagfXYZ
      stNormal;
   stNormal.fX = (float) ( stVector1.nY * stVector2.nZ - stVector1.nZ * stVector2.nY );
   stNormal.fY = (float) ( stVector1.nZ * stVector2.nX - stVector1.nX * stVector2.nZ );
   stNormal.fZ = (float) ( stVector1.nX * stVector2.nY - stVector1.nY * stVector2.nX );
   // normalize vector (make length 1)
   float
      fModulus = (float) sqrt ( stNormal.fX * stNormal.fX + stNormal.fY * stNormal.fY + stNormal.fZ * stNormal.fZ );
   poPol->m_stNormal.fX = stNormal.fX / fModulus;
   poPol->m_stNormal.fY = stNormal.fY / fModulus;
   poPol->m_stNormal.fZ = stNormal.fZ / fModulus;

   // compute plan D component
   poPol->m_fD = poPol->m_stNormal.fX * poVert0->m_stVertex.nX +
                 poPol->m_stNormal.fY * poVert0->m_stVertex.nY +
                 poPol->m_stNormal.fZ * poVert0->m_stVertex.nZ;
}



/* Rotate a point
**********************************************************/
void CMath::mRotate ( BYTE yAX, BYTE yAY, BYTE yAZ,
                      float& fX, float& fY, float& fZ )
{
   // rotated temp components
   float
      fXAux,
      fYAux,
      fZAux;

   fXAux = fX    * * ( m_pafCos + yAZ ) + fY    * * ( m_pafSin + yAZ );
   fYAux = fY    * * ( m_pafCos + yAZ ) - fX    * * ( m_pafSin + yAZ );
   fY    = fYAux * * ( m_pafCos + yAX ) + fZ    * * ( m_pafSin + yAX );
   fZAux = fZ    * * ( m_pafCos + yAX ) - fYAux * * ( m_pafSin + yAX );
   fX    = fXAux * * ( m_pafCos + yAY ) + fZAux * * ( m_pafSin + yAY );
   fZ    = fZAux * * ( m_pafCos + yAY ) - fXAux * * ( m_pafSin + yAY );
}
