/* Store data for a polygon mesh: vertexes in space, polygons, attributes
from polygons.

Vertexes: are locations in space (have 3 components: X, Y, Z)
Polygons: define index vertexes that compose one polygon
Attributes: define if polygon is a portal and your texture ID

WARNING:
- after change mesh data call CMesh::mBeginAddMesh and imediatly before
  call CMesh::mEndAddMesh.
*/



#include <string.h>
#include <math.h>

#include "defs.h"

#include "node.h"

#include "list.h"

#include "polygon.h"

#include "nodepol.h"

#include "nodeint.h"

#include "nodever.h"

#include "nodepor.h"

#include "map.h"

#include "math.h"

#include "mesh.h"



/* constructor - init empty sector
**********************************************************/
CMesh::CMesh ()
{
}



/* destructor - desalocate data, if alocated
**********************************************************/
CMesh::~CMesh ()
{
   mDestroy ();
}



/* Prepare mesh to reseve data: disable direct access feature.
return TRUE - ok / FALSE - error
Call after change mesh data, before call CMesh::mEndAddMesh
**********************************************************/
BOOL CMesh::mBeginAddMesh ()
{
   m_oPolygons.mDisableDirectAccess ();
   m_oVertexes.mDisableDirectAccess ();

   return TRUE;
}



/* Update mesh data acording polygons and vertexes definition:
recompute polygons normals and enable direct access to data lists.
return TRUE - ok / FALSE - error
Call after define all data and before use this data.
**********************************************************/
BOOL CMesh::mEndAddMesh ()
{
   // try enable direct access feature (speed up random access)
   if ( ! m_oPolygons.mEnableDirectAccess () ||
        ! m_oVertexes.mEnableDirectAccess () )
      return FALSE;   // if error ocurr return fail

   // compute polygons plan equation (used in collision detection,...)
   CMath
      oM;    // the math function is inside this object
   m_oPolygons.mSelectStart ();
   while ( m_oPolygons.mRetExistSelection () ) {

      // the current polygon
      CNodePolygon
         *poPol = (CNodePolygon*) m_oPolygons.mRetSelected ();
      
      oM.mDefinePolyPlanEquation ( m_oVertexes, poPol );

      // next polygon
      m_oPolygons.mSelectNext ();
   }

   return TRUE;
}



/* Desalocate mesh data, if alocated
**********************************************************/
void CMesh::mDestroy ()
{
   // delete array of indexes (inside each polygon struct)
   CNodePolygon
      *poPol = (CNodePolygon*) m_oPolygons.mRetStart ();
   while ( poPol != NULL ) {

      // delete array of indexes
      poPol->mDestroy ();

      // next polygon
      poPol = (CNodePolygon*) poPol->m_poNext;
   }

   // remove data
   m_oPolygons.mRemoveAll ();
   m_oVertexes.mRemoveAll ();
   m_oPortals.mRemoveAll ();
}



/* Set polygons attributes to mesh.
Obs.: first add polygons then set their attributes, else you can
have errors because no polygon defined to set attribute.
Convension: TRUE = 1, FALSE = 0
bPortal - TRUE the polygon is a portal;
yCrossable - flag if poly is crossable (see doc/TXE-spec.txt)
dwTextureID - texture ID for the polygon;
nNdxPol1 - first polygon to apply attribute;
nNdxPol2 - last polygon to apply attribute;
Polygons that have indexes between nNdxPol1 and nNdxPol2
will have their attributes changed.
return TRUE - ok / FALSE - indexes out of range
Obs.: after change mesh data call CMesh::mBeginAddMesh.
**********************************************************/
BOOL CMesh::mSetPolygonsAttributes ( BOOL bPortal, BYTE yCrossable, DWORD dwTextureID, INT nNdxPol1, INT nNdxPol2 )
{
   // sort indexes
   INT
      nMin = nNdxPol1,
      nMax = nNdxPol2;
   if ( nNdxPol1 > nNdxPol2 ) {
      nMin = nNdxPol2,
      nMax = nNdxPol1;
   }

   // verify if polygons indexes in sector are out of range, if yes return
   if ( nMin < 0 || nMax >= (INT) m_oPolygons.mRetNumNodes () ) {
      printf ( MSG_ERROR_INDEX_OUT_OF_RANGE );
      return FALSE;
   }

   // set attribute values
   for ( INT ncI = nMin; ncI <= nMax; ncI ++ ) {
      ( (CNodePolygon*) m_oPolygons [ ncI ] )->m_bPortal = bPortal;
      ( (CNodePolygon*) m_oPolygons [ ncI ] )->m_yCrossable = yCrossable;
      ( (CNodePolygon*) m_oPolygons [ ncI ] )->m_dwTextureID = dwTextureID;
   }

   return TRUE;
}



/* Set polygons light intensity (level)
Convension: TRUE = 1, FALSE = 0
fLight can be between 0 and 1, is the light level (0 none, 1 maximum)
dwLightID define what light table use (they have one ID defined in loadtime)
nNdxPol1 - first polygon to apply attribute;
nNdxPol2 - last polygon to apply attribute;
Polygons that have indexes between nNdxPol1 and nNdxPol2
will have their attributes changed.
return TRUE - ok / FALSE - indexes out of range
Obs.: after change mesh data call CMesh::mBeginAddMesh.
**********************************************************/
BOOL CMesh::mSetPolygonsLight ( float fLight, DWORD dwLightID,
                                INT nNdxPol1, INT nNdxPol2 )
{
   // sort indexes
   INT
      nMin = nNdxPol1,
      nMax = nNdxPol2;
   if ( nNdxPol1 > nNdxPol2 ) {
      nMin = nNdxPol2,
      nMax = nNdxPol1;
   }

   // verify if polygons indexes in sector are out of range, if yes return
   if ( nMin < 0 || nMax >= (INT) m_oPolygons.mRetNumNodes () ) {
      printf ( MSG_ERROR_INDEX_OUT_OF_RANGE );
      return FALSE;
   }

   // set attribute value
   for ( INT ncI = nMin; ncI <= nMax; ncI ++ ) {
      ( (CNodePolygon*) m_oPolygons [ ncI ] )->m_fLight = fLight;
      ( (CNodePolygon*) m_oPolygons [ ncI ] )->m_dwLightID = dwLightID;
   }

   return TRUE;
}



/* 1 - Add polygons
oIndexes is the list with vertex indexes that compose the polygon,
the indexes are acording vertex list in mesh.
return TRUE - ok \ FALSE - error, no memory
Obs.: after change mesh data call CMesh::mBeginAddMesh.
**********************************************************/
BOOL CMesh::mAddPolygon ( CList& oIndexes )
{
   // try create new polygon
   CNodePolygon
      *poPol = new CNodePolygon;
   if ( poPol == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // create a new polygon
   poPol->mCreate ( oIndexes );

   // update mesh adding polygon and attribute
   m_oPolygons.mSetEnd ( (CNode*) poPol );

   return TRUE;
}



/* 2- Add polygons
Make a copy from the source polygon poSrc.
return TRUE - ok \ FALSE - error, no memory
Obs.: after change mesh data call CMesh::mBeginAddMesh.
**********************************************************/
BOOL CMesh::mAddPolygon ( CNodePolygon* poSrc )
{
   // try create new polygon
   CNodePolygon
      *poPol = new CNodePolygon;
   if ( poPol == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // copy
   poPol->mSetObj ( *poSrc );

   // update mesh adding polygon
   m_oPolygons.mSetEnd ( (CNode*) poPol );

   return TRUE;
}



/* Add vertexes coordenates.
bScale - TRUE - scale vertex (vertex * power(2,SCALEEXP))
return TRUE - ok \ FALSE - error, no memory
Obs.: after change mesh data call CMesh::mBeginAddMesh.
**********************************************************/
BOOL CMesh::mAddVertex ( INT nX, INT nY, INT nZ, BOOL bScale )
{
   // try create new vertex
   CNodeVertex
      *poVert = new CNodeVertex;
   if ( poVert == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // define vertex data
   // if need scale vertex
   if ( bScale ) {
      poVert->m_stVertex.nX = nX << SCALEEXP;  // == nX * power(2,SCALEEXP)
      poVert->m_stVertex.nY = nY << SCALEEXP;
      poVert->m_stVertex.nZ = nZ << SCALEEXP;
   }
   else {
      poVert->m_stVertex.nX = nX;
      poVert->m_stVertex.nY = nY;
      poVert->m_stVertex.nZ = nZ;
   }

   // update data
   m_oVertexes.mSetEnd ( (CNode*) poVert );

   return TRUE;
}



/* Add a portal polygon definition. between 2 sectors.
nNdxPol is the index, in polygon list, of the polygon that
is portal (the polygon list is in mesh data).
dwObjectID is the ID of destin object where sector is.
dwSectorID is the destin sector ID.
return TRUE - ok \ FALSE - error, no memory
Obs.: after change data call CSector::mBeginAddSector.
**********************************************************/
BOOL CMesh::mAddPortal ( DWORD dwNdxPol, DWORD dwObjectID, DWORD dwSectorID )
{
   // try create new portal node
   CNodePortal
      *poDestin = new CNodePortal;
   if ( poDestin == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // define portal data, destin sector
   poDestin->m_stPortal.dwObjectID = dwObjectID;
   poDestin->m_stPortal.dwSectorID = dwSectorID;

   // update sector adding portal
   m_oPortals.mAddKey ( dwNdxPol, (CNode*) poDestin );

   return TRUE;
}
