/* Store all data needed in a scenery: CObject instances, textures, the
palette (if needed), cameras for visualization.

TIPS:
- the data can be loaded from a TXE file, see class CFileTXE for more info.

WARNING:
- after change object or sector data you MUST CALL CScenery::mBeginAddObject
  or CScenery::mBeginAddSector, respectively.
- before change object or sector data you MUST CALL CScenery::mEndAddObject
  or CScenery::mEndAddSector, respectively.
- Use CScenery::mSetCurrentCamera to change the current camera
*/



#include <string.h>

#include "defs.h"

#include "mem.h"

#include "string.h"

#include "file.h"

#include "filerem.h"

#include "filepcx.h"

#include "node.h"

#include "list.h"

#include "polygon.h"

#include "nodepol.h"

#include "map.h"

#include "texture.h"

#include "lighttab.h"

#include "nodelgt.h"

#include "nodepor.h"

#include "mesh.h"

#include "sector.h"

#include "nodesec.h"

#include "math.h"

#include "refsys.h"

#include "camsys.h"

#include "object.h"

#include "nodeint.h"

#include "nodetex.h"

#include "nodeobj.h"

#include "nodever.h"

#include "filetxe.h"

#include "scenery.h"



/* constructor - init empty scenery
**********************************************************/
CScenery::CScenery ()
{
   // flag object definition
   m_bDefiningObject = TRUE;

   m_dwSectorID = m_dwObjectID = 0;
   m_nClearBackground = -1;
   m_dwCameraID = (DWORD) -1;
   m_dwInsideObjID = m_dwInsideSecID = 0;
}



/* constructor - load TXE file to this scenery,
if fail abort programm
**********************************************************/
CScenery::CScenery ( CString oszFile )
{
   // flag object definition
   m_bDefiningObject = TRUE;

   m_dwSectorID = m_dwObjectID = 0;
   m_nClearBackground = -1;
   m_dwCameraID = (DWORD) -1;
   m_dwInsideObjID = m_dwInsideSecID = 0;

   // try load
   if ( ! mCreate ( oszFile ) )
      ABORT ( MSG_FILE_NOT_FOUND, __FILE__, __LINE__ )
}



/* destructor - desalocate buffer, if alocated
**********************************************************/
CScenery::~CScenery ()
{
   mDestroy ();
}



/* Load scenery data. (like CScenery::constructor)
return TRUE- ok / FALSE - error
**********************************************************/
BOOL CScenery::mCreate ( CString oszFile )
{
   mDestroy ();

   // try load
   CFileTXE
      oFile;
   if ( ! oFile.mLoad ( oszFile, *this ) )
      return FALSE;

   return TRUE;
}



/* Empty scenery data.
**********************************************************/
void CScenery::mDestroy ()
{
   // flag object definition
   m_bDefiningObject = TRUE;

   m_dwSectorID = m_dwObjectID = 0;
   m_nClearBackground = -1;
   m_dwCameraID = (DWORD) -1;
   m_dwInsideObjID = m_dwInsideSecID = 0;

   m_oPal.mDestroy ();
   m_oTextures.mRemoveAll ();
   m_oObjects.mRemoveAll ();
   m_oLightTables.mRemoveAll ();
}



/* Set initial engine parameters.
dwCameraID - what instance is the camera (get keyboard control)
dwObjID - position of camera (where camera is), inside what instance
dwSecID - inside what sector in instance
nClear - background color or -1 if no frame buffer clear is need
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CScenery::mSetStart ( DWORD dwCameraID, DWORD dwObjID,
                           DWORD dwSecID, INT nClear )
{
   // define frame buffer clear color, used as background color (-1 for none)
   m_nClearBackground = nClear;

   // what instance have the camera control
   return mSetCurrentCamera ( dwCameraID, dwObjID, dwSecID );
}   



/* Set the object as the current camera.
Invert the vector K component Z.
return TRUE- ok / FALSE - error
**********************************************************/
BOOL CScenery::mSetCurrentCamera ( DWORD dwCameraID, DWORD dwObjID,
                                   DWORD dwSecID )
{
   // invert some camera components of the last camera, if exist
   if ( m_dwCameraID != (DWORD) -1 ) {
      // search for object
      CNodeObject
         *poObj = (CNodeObject*) m_oRootInstances.mSearchObject ( m_dwCameraID );
      // verify if object not found
      if ( poObj == NULL ) {
         printf ( MSG_ERROR_CAMERA_OBJECTID_NOT_FOUND, dwObjID );
         printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
         return FALSE;
      }
      mInvertCamera ( poObj->m_oRefSys );
   }
   
   // the new camera
   m_dwCameraID = dwCameraID;

   // update camera
   // search for object
   CNodeObject
      *poObj = (CNodeObject*) m_oRootInstances.mSearchObject ( m_dwCameraID );
   // verify if object not found
   if ( poObj == NULL ) {
      printf ( MSG_ERROR_CAMERA_OBJECTID_NOT_FOUND, dwObjID );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }
   mInvertCamera ( poObj->m_oRefSys );

   // where the camera start
   m_dwInsideObjID = dwObjID;
   m_dwInsideSecID = dwSecID;

   return TRUE;
}



/* Invert camera to use it as the view camera system or as
an object RS.
oCamSys the given camera to invert.
**********************************************************/
void CScenery::mInvertCamera ( CCameraSystem& oCamSys )
{
   // invert K
   oCamSys.m_stK.fX = - oCamSys.m_stK.fX;
   oCamSys.m_stK.fY = - oCamSys.m_stK.fY;
   oCamSys.m_stK.fZ = - oCamSys.m_stK.fZ;
   // update base
   oCamSys.mComputeFixPointBase ();
   // update position
   oCamSys.m_stPos.fX = - oCamSys.m_stPos.fX;
   oCamSys.m_stPos.fY = - oCamSys.m_stPos.fY;
   oCamSys.m_stPos.fZ = - oCamSys.m_stPos.fZ;
}



/* Add palette for graphics mode (if they have)
return TRUE - ok / FALSE - error (palette already defined)
**********************************************************/
BOOL CScenery::mAddPalette ( CMem oPal )
{
   // verify if palette have been already defined
   if ( m_oPal.m_dwSize != 0 ) {
      printf ( MSG_ERROR_PALETTE_ALREADY_DEFINED );
      return FALSE;
   }

   // copy palette from source oPal to m_oPal
   m_oPal = oPal;

   return TRUE;
}



/* Add light table for shaded polys.
If oLgT is empty, create light table from source palette
(CScenery::m_oPal) using dwNumLevels and save it to file oszFile.
dwID - light table ID
fR, fG, fB - light source color (if need), the values are between 0 and 1;
return TRUE - ok / FALSE - error (impossible define light table:
if is already defined or creation error - no source palette)
**********************************************************/
BOOL CScenery::mAddLightTable ( CLightTable& oLgT, DWORD dwID, CString oszFile,
                                DWORD dwNumLevels, float fR, float fG, float fB )
{
   // if key already defined, return
   if ( m_oLightTables [ dwID ] != NULL ) {
      printf ( MSG_ERROR_LIGHTTABLE_ID_ALREADY_DEFINED, dwID );
      return FALSE;
   }

   // verify if loaded light table have error, if yes try create it
   if ( oLgT.m_payLightTable == NULL ) {

      // verify if source palette have been already defined
      if ( m_oPal.m_dwSize == 0 ) {
         printf ( MSG_ERROR_NOPALETTE_FOR_LIGHTTABLE );
         return FALSE;
      }
     
      // create it
      oLgT.mCreate ( m_oPal, dwNumLevels, fR, fG, fB ); 

      // save light table
      oLgT.mSave ( oszFile );
   }

   // try create light table node
   CNodeLightTable
      *poLgT = new CNodeLightTable ( oLgT );
   if ( poLgT == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // add in scenery
   if ( ! m_oLightTables.mAddKey ( dwID, (CNode*) poLgT ) ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   return TRUE;
}
   


/* Add texture to map relation. Map relations are implemented
by CMap class and define a relationship between a key and one value,
here the key is the texture ID and the value is the texture.
See CMap implementation file for more info.
return TRUE - ok / FALSE - error (texture ID already defined)
**********************************************************/
BOOL CScenery::mAddTexture ( DWORD dwID, CTexture& oTx )
{
   // if key already defined, return
   if ( m_oTextures [ dwID ] != NULL ) {
      printf ( MSG_ERROR_TEXTURE_ID_ALREADY_DEFINED, dwID );
      return FALSE;
   }

   // try create texture node
   CNodeTexture
      *poTx = new CNodeTexture ( oTx );
   if ( poTx == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // add in scenery
   if ( ! m_oTextures.mAddKey ( dwID, (CNode*) poTx ) ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   return TRUE;
}



/* Add an object instance in scenery.
Before call this function add all object models in scenery.
wInstID - this instance ID (must be uniq in ALL scenery)
dwObjectID - the object model ID used for this instance
dwDestinInstID - inside of what instanciated object insert this instance
dwDestinSectorID - inside of what sector (in object) insert this instance
stPos - the initial position in space (SRU)
stI, stJ, stK - the initial reference system base (SRU)
The initial values of base and position are acording the base and position
of the universe reference system.
bCopyData - TRUE - this instance have your own mesh / FALSE use the
            shared model mesh (good for memory safe and animation)
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CScenery::mAddInstance ( DWORD dwInstID, DWORD dwObjectID,
                              DWORD dwDestinInstID,DWORD dwDestinSectorID,
                              tagXYZ& stPos, tagfXYZ& stI,
                              tagfXYZ& stJ, tagfXYZ& stK,
                              BOOL bCopyData ) // REVIEW bCopyData nao usado, sempre copia
{
   // verify if the given instance is already defined
   if ( m_oRootInstances.mSearchObject ( dwInstID ) != NULL ) {
      printf ( MSG_ERROR_OBJECT_ID_ALREADY_DEFINED, dwInstID );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // search sector where insert instance
   CNodeObject
      *poObj;
   CNodeSector
      *poSec;
   // verify if insert instance in root sector (first scenery sector)
   if ( (INT) dwDestinInstID == -1 )
      // use the root sector
      poSec = (CNodeSector*) &m_oRootInstances;
   else {
      // if the object is defined
      poObj = m_oRootInstances.mSearchObject ( dwDestinInstID );
      // verify if not found -> error
      if ( poObj == NULL ) {
         printf ( MSG_ERROR_DESTIN_OBJECT_NOT_FOUND, dwDestinInstID );
         printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
         return FALSE;
      }
      poSec = (CNodeSector*) poObj->m_oSectors.mRetValue ( dwDestinSectorID );
      // verify if not found -> error
      if ( poSec == NULL ) {
         printf ( MSG_ERROR_DESTIN_SECTOR_NOT_FOUND, dwDestinSectorID );
         printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
         return FALSE;
      }
   }

   // define the source model
   CNodeObject
      *poModel = (CNodeObject*) m_oObjects.mRetValue ( dwObjectID );
   // verify if not found -> error
   if ( poModel == NULL ) {
      printf ( MSG_ERROR_SOURCE_MODEL_NOT_FOUND, dwObjectID );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // create the instance
   CNodeObject
      *poNewObj = new CNodeObject;
   // verify error
   if ( poNewObj == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // start copy
   poNewObj->mBeginAddObject ();

   // copy the object mesh
   // REVIEW por toda a copia (pols, vertices, remover portais - por
   // dentro dos poligonos) numa funcao de CMesh

   // copy the polygons from the source to new
   poModel->m_oPolygons.mSelectStart ();
   while ( poModel->m_oPolygons.mRetExistSelection () ) {
      poNewObj->mAddPolygon ( (CNodePolygon*) poModel->m_oPolygons.mRetSelected () );
      poModel->m_oPolygons.mSelectNext ();
   }

   // copy the vertexes from the source to new
   poModel->m_oVertexes.mSelectStart ();
   while ( poModel->m_oVertexes.mRetExistSelection () ) {
      poNewObj->mAddVertex (
      ( (CNodeVertex*) poModel->m_oVertexes.mRetSelected () ) -> m_stVertex.nX,
      ( (CNodeVertex*) poModel->m_oVertexes.mRetSelected () ) -> m_stVertex.nY,
      ( (CNodeVertex*) poModel->m_oVertexes.mRetSelected () ) -> m_stVertex.nZ,
      FALSE   // no scale (model is already scaled)
      );
      poModel->m_oVertexes.mSelectNext ();
   }

   // copy the portal definition from source to new
   for ( DWORD dwcK = 0; dwcK < poModel->m_oPortals.mRetSize (); dwcK ++ ) {
      // get key (polygon portal index)
      DWORD
         dwKey = poModel->m_oPortals.mRetKey ( dwcK );
      // get value (portal destin)
      CNodePortal
         *poPortal = (CNodePortal*) poModel->m_oPortals.mRetValue ( dwKey );
      // add
      poNewObj->mAddPortal ( dwKey, poPortal->m_stPortal.dwObjectID,
                                    poPortal->m_stPortal.dwSectorID );
   }

   // copy the sectors mesh

   // for each sector in object
   for ( DWORD dwcS = 0; dwcS < poModel->m_oSectors.mRetSize (); dwcS ++ ) {

      // get the source sector ID
      DWORD
         dwSecID = poModel->m_oSectors.mRetKey ( dwcS );
      // get the source sector
      CNodeSector
         *poSrcSec = (CNodeSector*) poModel->m_oSectors.mRetValue ( dwSecID );

      // create and add sector
      CNodeSector
         *poNewSec = poNewObj->mAddSector ( dwSecID, dwInstID );

      // add mesh to sector
      poNewSec->mBeginAddSector ();

      // copy the polygons from the source to new
      poSrcSec->m_oPolygons.mSelectStart ();
      while ( poSrcSec->m_oPolygons.mRetExistSelection () ) {
         poNewSec->mAddPolygon ( (CNodePolygon*) poSrcSec->m_oPolygons.mRetSelected () );
         poSrcSec->m_oPolygons.mSelectNext ();
      }

      // copy the vertexes from the source to new
      poSrcSec->m_oVertexes.mSelectStart ();
      while ( poSrcSec->m_oVertexes.mRetExistSelection () ) {
         poNewSec->mAddVertex (
         ( (CNodeVertex*) poSrcSec->m_oVertexes.mRetSelected () ) -> m_stVertex.nX,
         ( (CNodeVertex*) poSrcSec->m_oVertexes.mRetSelected () ) -> m_stVertex.nY,
         ( (CNodeVertex*) poSrcSec->m_oVertexes.mRetSelected () ) -> m_stVertex.nZ,
         FALSE   // no scale (model is already scaled)
         );

         poSrcSec->m_oVertexes.mSelectNext ();
      }

      // copy the portal definition from source to new
      for ( DWORD dwcK = 0; dwcK < poSrcSec->m_oPortals.mRetSize (); dwcK ++ ) {
         // get key (polygon portal index)
         DWORD
            dwKey = poSrcSec->m_oPortals.mRetKey ( dwcK );
         // get value (portal destin)
         CNodePortal
            *poPortal = (CNodePortal*) poSrcSec->m_oPortals.mRetValue ( dwKey );

         // add
         poNewSec->mAddPortal ( dwKey, poPortal->m_stPortal.dwObjectID,
                                       poPortal->m_stPortal.dwSectorID );
      }

      // end add mesh to sector
      poNewSec->mEndAddSector ();
   }

   // define object reference system parameters (position and orientation)
   // position
   // REVIEW -> os - abaixo podem causar problemas em outros SR?
   poNewObj->m_oRefSys.m_stPos.fX = - ( stPos.nX << SCALEEXP ); // == nX * power(2,SCALEEXP)
   poNewObj->m_oRefSys.m_stPos.fY = - ( stPos.nY << SCALEEXP );
   poNewObj->m_oRefSys.m_stPos.fZ = stPos.nZ << SCALEEXP;
   // orientation
   poNewObj->m_oRefSys.m_stI.fX = stI.fX;
   poNewObj->m_oRefSys.m_stI.fY = stI.fY;
   poNewObj->m_oRefSys.m_stI.fZ = stI.fZ;
   poNewObj->m_oRefSys.m_stJ.fX = stJ.fX;
   poNewObj->m_oRefSys.m_stJ.fY = stJ.fY;
   poNewObj->m_oRefSys.m_stJ.fZ = stJ.fZ;
   poNewObj->m_oRefSys.m_stK.fX = stK.fX;
   poNewObj->m_oRefSys.m_stK.fY = stK.fY;
   poNewObj->m_oRefSys.m_stK.fZ = stK.fZ;
   // define fix point base
   poNewObj->m_oRefSys.mComputeFixPointBase ();

   // end object creation
   poNewObj->mEndAddObject ();

   // add created instance in respective sector
   poSec->mBeginAddSector ();
   poSec->m_oInstances.mAddKey ( dwInstID, (CNode*) poNewObj );
   poSec->mEndAddSector ();

   return TRUE;
}



/* Move an object instance between sectors.
Change current camera object and sector if the object
moved is the camera.
**********************************************************/
void CScenery::mMoveInstance ( CNodeObject& oObj, DWORD dwObjID,
                               CNodeSector& oParentSec,
                               DWORD dwDestinObjID, DWORD dwDestinSecID )

{
   // remove the sector references for the object
   oParentSec.m_oInstances.mRemoveKey ( dwObjID, FALSE );

   // search new sector
   // REVIEW otimizar: procurar nos objetos dentro do antigo
   // setor, nos setores adjacentes ao antigo setor ou no setor
   // pai do antigo setor.
   CNodeObject
      *poNewParentObj = (CNodeObject*) m_oRootInstances.mSearchObject ( dwDestinObjID );
   CNodeSector
      *poNewParentSec = (CNodeSector*) poNewParentObj->m_oSectors.mRetValue ( dwDestinSecID );

   // add references to new sector
   poNewParentSec->m_oInstances.mAddKey ( dwObjID, (CNode*) &oObj );

   // change current object and sector if is the camera
   if ( dwObjID == m_dwCameraID ) {
      m_dwInsideObjID = dwDestinObjID;
      m_dwInsideSecID = dwDestinSecID;
   }
}



/* Start definition of a new object in scenery
dwObjID is the object ID
return TRUE - ok / FALSE - error (ID already defined)
Obs.: before call CScenery::mEndAddObject.
**********************************************************/
BOOL CScenery::mBeginAddObject ( DWORD dwObjID )
{
   // if key already defined, ID exist, return
   if ( m_oObjects [ dwObjID ] != NULL ) {
      printf ( MSG_ERROR_OBJECT_ID_ALREADY_DEFINED, dwObjID );
      return FALSE;
   }

   // try create object node
   CNodeObject
      *poObj = new CNodeObject;
   if ( poObj == NULL ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // try define start of object data update
   if ( ! poObj->mBeginAddObject () )
      return FALSE;

   // add in scenery
   if ( ! m_oObjects.mAddKey ( dwObjID, (CNode*) poObj ) ) {
      printf ( MSG_ERROR_MEMORY_LOW );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // flag object definition
   m_bDefiningObject = TRUE;
   m_dwObjectID = dwObjID;       // object ID

   return TRUE;
}



/* Flag end of object definition. Call after define all
object data and before use this data.
return TRUE - ok \ FALSE - error
**********************************************************/
BOOL CScenery::mEndAddObject ()
{
   // update object data
   // point object
   CNodeObject
      *poObj = (CNodeObject*) m_oObjects [ m_dwObjectID ];
   // do update
   return poObj->mEndAddObject ();
}



/* Start definition of a new sector in current object.
dwSectorID is the sector ID
return TRUE - ok / FALSE - error (ID already defined,...)
Obs.: before call CScenery::mEndAddSector.
**********************************************************/
BOOL CScenery::mBeginAddSector ( DWORD dwSectorID )
{
   // object where sector will be inserted
   CNodeObject
      *poObj = (CNodeObject*) m_oObjects [ m_dwObjectID ];
   // if point to NULL: object ID not found
   if ( poObj == NULL ) {
      printf ( MSG_ERROR_ID_NOT_FOUND );
      printf ( MSG_DEFAULT_ABORT_INFO, __FILE__, __LINE__ );
      return FALSE;
   }

   // add in object and prepare sector to reseve data
   CNodeSector
      *poNew = poObj->mAddSector ( dwSectorID, m_dwObjectID );
   // if error
   if ( poNew == NULL )
      return FALSE;

   // try start sector data definition
   if ( ! poNew->mBeginAddSector () )
      return FALSE;

   // flag sector definition
   m_bDefiningObject = FALSE;
   m_dwSectorID = dwSectorID;

   return TRUE;
}



/* Flag end of more sector definition, now is object definition.
Call after define all data and before use this data.
return TRUE - ok \ FALSE - error
**********************************************************/
BOOL CScenery::mEndAddSector ()
{
   // flag object definition
   m_bDefiningObject = TRUE;

   // update sector data
   // object where sector is
   CNodeObject
      *poObj = (CNodeObject*) m_oObjects [ m_dwObjectID ];
   // point the sector
   CNodeSector
      *poSector = (CNodeSector*) poObj->m_oSectors [ m_dwSectorID ];
   return poSector->mEndAddSector ();
}



/* Set polygons attributes to object or sector.
Convension: TRUE = 1, FALSE = 0
bPortal - TRUE the polygon is a portal;
yCrossable - flag if poly is crossable (see doc/TXE-spec.txt)
dwTextureID - texture ID for the polygon;
nNdxPol1 - first polygon to apply attribute;
nNdxPol2 - last polygon to apply attribute;
Polygons that have indexes between nNdxPol1 and nNdxPol2
will have their attributes changed.
return TRUE - ok / FALSE - indexes out of range
Obs.: after change data call CScenery::mBeginAddObject or
 CScenery::mBeginAddSector.
**********************************************************/
BOOL CScenery::mSetPolygonsAttributes ( BOOL bPortal, BYTE yCrossable, DWORD dwTextureID, INT nNdxPol1, INT nNdxPol2 )
{
   // verify if include in object
   if ( m_bDefiningObject ) {
      // define mesh to set attribute
      // object
      CNodeObject
         *poObj = (CNodeObject*) m_oObjects [ m_dwObjectID ];
      // set attributes
      return poObj->mSetPolygonsAttributes ( bPortal, yCrossable, dwTextureID, nNdxPol1, nNdxPol2 );
   }

   // define mesh to set attribute
   // object where sector is
   CNodeObject
      *poObj = (CNodeObject*) m_oObjects [ m_dwObjectID ];
   // point the sector
   CNodeSector
      *poSector = (CNodeSector*) poObj->m_oSectors [ m_dwSectorID ];
   // set attributes
   return poSector->mSetPolygonsAttributes ( bPortal, yCrossable, dwTextureID, nNdxPol1, nNdxPol2 );
}



/* Set polygons attributes to object or sector.
fLight is a value between 0 and 1.
dwLightID - light source (light table ID)
nNdxPol1 - first polygon to apply attribute;
nNdxPol2 - last polygon to apply attribute;
Polygons that have indexes between nNdxPol1 and nNdxPol2
will have their attributes changed.
return TRUE - ok / FALSE - indexes out of range
Obs.: after change data call CScenery::mBeginAddObject or
 CScenery::mBeginAddSector.
**********************************************************/
BOOL CScenery::mSetPolygonsLight ( float fLight, DWORD dwLightID,
                                   INT nNdxPol1, INT nNdxPol2 )
{
   // verify if include in object
   if ( m_bDefiningObject ) {
      // define mesh to set attribute
      // object
      CNodeObject
         *poObj = (CNodeObject*) m_oObjects [ m_dwObjectID ];
      // set attributes
      return poObj->mSetPolygonsLight ( fLight, dwLightID, nNdxPol1, nNdxPol2 );
   }

   // define mesh to set attribute
   // object where sector is
   CNodeObject
      *poObj = (CNodeObject*) m_oObjects [ m_dwObjectID ];
   // point the sector
   CNodeSector
      *poSector = (CNodeSector*) poObj->m_oSectors [ m_dwSectorID ];
   // set attributes
   return poSector->mSetPolygonsLight ( fLight, dwLightID, nNdxPol1, nNdxPol2 );
}



/* Add polygons in current scenery object or sector.
oIndexes is the list with vertex indexes that compose the polygon,
the indexes are acording vertex list in mesh.
return TRUE - ok \ FALSE - error, no memory
Obs.: after change data call CScenery::mBeginAddObject or
 CScenery::mBeginAddSector.
**********************************************************/
BOOL CScenery::mAddPolygon ( CList& oIndexes )
{
   // verify if include in object
   if ( m_bDefiningObject ) {
      // define mesh to set attribute
      // object
      CNodeObject
         *poObj = (CNodeObject*) m_oObjects [ m_dwObjectID ];
      // set attributes
      return poObj->mAddPolygon ( oIndexes );
   }

   // update sector
   // object where sector is
   CNodeObject
      *poObj = (CNodeObject*) m_oObjects [ m_dwObjectID ];
   // point the sector
   CNodeSector
      *poSector = (CNodeSector*) poObj->m_oSectors [ m_dwSectorID ];
   // add polygon
   return poSector->mAddPolygon ( oIndexes );
}



/* Add vertexes coordenates in a object or sector
bScale define if need scale de vertex (TRUE - yes / FALSE - no)
return TRUE - ok \ FALSE - error, no memory
Obs.: after change data call CScenery::mBeginAddObject or
 CScenery::mBeginAddSector.
**********************************************************/
BOOL CScenery::mAddVertex ( INT nX, INT nY, INT nZ, BOOL bScale )
{
   // verify if include in object
   if ( m_bDefiningObject ) {
      // define mesh to set attribute
      // object
      CNodeObject
         *poObj = (CNodeObject*) m_oObjects [ m_dwObjectID ];
      // set attributes
      return poObj->mAddVertex ( nX, nY, nZ, bScale );
   }

   // update sector
   // object where sector is
   CNodeObject
      *poObj = (CNodeObject*) m_oObjects [ m_dwObjectID ];
   // point the sector
   CNodeSector
      *poSector = (CNodeSector*) poObj->m_oSectors [ m_dwSectorID ];
   // add vertex
   return poSector->mAddVertex ( nX, nY, nZ, bScale );
}



/* Add a portal between 2 sectors.
nNdxPol is the index of polygon that is portal.
dwSectorID is the destin sector.
return TRUE - ok \ FALSE - error, no memory
Obs.: after change data call CScenery::mBeginAddObject or
 CScenery::mBeginAddSector.
**********************************************************/
BOOL CScenery::mAddPortal ( DWORD dwNdxPol, DWORD dwObjectID, DWORD dwSectorID )
{
   // verify if must include in object, if yes warns: no object inclusion now
   if ( m_bDefiningObject ) {
      // define mesh to set attribute
      // object
      CNodeObject
         *poObj = (CNodeObject*) m_oObjects [ m_dwObjectID ];
      // set attributes
      return poObj->mAddPortal ( dwNdxPol, dwObjectID, dwSectorID );
   }

   // update sector
   // object where sector is
   CNodeObject
      *poObj = (CNodeObject*) m_oObjects [ m_dwObjectID ];
   // point the sector
   CNodeSector
      *poSector = (CNodeSector*) poObj->m_oSectors [ m_dwSectorID ];
   // add portal
   return poSector->mAddPortal ( dwNdxPol, dwObjectID, dwSectorID );
}
